/* This is a generated file by dump_symbols.h. *DO NOT EDIT MANUALLY !* */
#define adler32 gdal_adler32
#define adler32_combine gdal_adler32_combine
#define adler32_combine_ gdal_adler32_combine_
#define adler32_combine64 gdal_adler32_combine64
#define adler32_z gdal_adler32_z
#define bi_flush gdal_bi_flush
#define bi_reverse gdal_bi_reverse
#define bi_windup gdal_bi_windup
#define build_bl_tree gdal_build_bl_tree
#define build_tree gdal_build_tree
#define byte_swap gdal_byte_swap
#define compress gdal_compress
#define compress2 gdal_compress2
#define compress_block gdal_compress_block
#define compressBound gdal_compressBound
#define crc32 gdal_crc32
#define crc32_combine gdal_crc32_combine
#define crc32_combine64 gdal_crc32_combine64
#define crc32_combine_gen gdal_crc32_combine_gen
#define crc32_combine_gen64 gdal_crc32_combine_gen64
#define crc32_combine_op gdal_crc32_combine_op
#define crc32_z gdal_crc32_z
#define crc_word gdal_crc_word
#define crc_word_big gdal_crc_word_big
#define deflate gdal_deflate
#define deflateBound gdal_deflateBound
#define deflateCopy gdal_deflateCopy
#define deflateEnd gdal_deflateEnd
#define deflate_fast gdal_deflate_fast
#define deflateGetDictionary gdal_deflateGetDictionary
#define deflate_huff gdal_deflate_huff
#define deflateInit_ gdal_deflateInit_
#define deflateInit2_ gdal_deflateInit2_
#define deflateParams gdal_deflateParams
#define deflatePending gdal_deflatePending
#define deflatePrime gdal_deflatePrime
#define deflateReset gdal_deflateReset
#define deflateResetKeep gdal_deflateResetKeep
#define deflate_rle gdal_deflate_rle
#define deflateSetDictionary gdal_deflateSetDictionary
#define deflateSetHeader gdal_deflateSetHeader
#define deflate_slow gdal_deflate_slow
#define deflateStateCheck gdal_deflateStateCheck
#define deflate_stored gdal_deflate_stored
#define deflateTune gdal_deflateTune
#define detect_data_type gdal_detect_data_type
#define fill_window gdal_fill_window
#define fixedtables gdal_fixedtables
#define fixedtables gdal_fixedtables
#define flush_pending gdal_flush_pending
#define frame_dummy gdal_frame_dummy
#define gen_bitlen gdal_gen_bitlen
#define gen_codes gdal_gen_codes
#define get_crc_table gdal_get_crc_table
#define inflate gdal_inflate
#define inflateBack gdal_inflateBack
#define inflateBackEnd gdal_inflateBackEnd
#define inflateBackInit_ gdal_inflateBackInit_
#define inflateCodesUsed gdal_inflateCodesUsed
#define inflateCopy gdal_inflateCopy
#define inflateEnd gdal_inflateEnd
#define inflate_fast gdal_inflate_fast
#define inflateGetDictionary gdal_inflateGetDictionary
#define inflateGetHeader gdal_inflateGetHeader
#define inflateInit_ gdal_inflateInit_
#define inflateInit2_ gdal_inflateInit2_
#define inflateMark gdal_inflateMark
#define inflatePrime gdal_inflatePrime
#define inflateReset gdal_inflateReset
#define inflateReset2 gdal_inflateReset2
#define inflateResetKeep gdal_inflateResetKeep
#define inflateSetDictionary gdal_inflateSetDictionary
#define inflateStateCheck gdal_inflateStateCheck
#define inflateSync gdal_inflateSync
#define inflateSyncPoint gdal_inflateSyncPoint
#define inflate_table gdal_inflate_table
#define inflateUndermine gdal_inflateUndermine
#define inflateValidate gdal_inflateValidate
#define init_block gdal_init_block
#define lm_init gdal_lm_init
#define longest_match gdal_longest_match
#define multmodp gdal_multmodp
#define pqdownheap gdal_pqdownheap
#define putShortMSB gdal_putShortMSB
#define read_buf gdal_read_buf
#define scan_tree gdal_scan_tree
#define send_all_trees gdal_send_all_trees
#define send_tree gdal_send_tree
#define slide_hash gdal_slide_hash
#define syncsearch gdal_syncsearch
#define _tr_align gdal__tr_align
#define _tr_flush_bits gdal__tr_flush_bits
#define _tr_flush_block gdal__tr_flush_block
#define _tr_init gdal__tr_init
#define tr_static_init gdal_tr_static_init
#define _tr_stored_block gdal__tr_stored_block
#define _tr_tally gdal__tr_tally
#define uncompress gdal_uncompress
#define uncompress2 gdal_uncompress2
#define updatewindow gdal_updatewindow
#define x2nmodp gdal_x2nmodp
#define zcalloc gdal_zcalloc
#define zcfree gdal_zcfree
#define zError gdal_zError
#define zlibCompileFlags gdal_zlibCompileFlags
#define zlibVersion gdal_zlibVersion
#define crc_table gdal_crc_table
#define crc_big_table gdal_crc_big_table
#define crc_braid_table gdal_crc_braid_table
#define crc_braid_big_table gdal_crc_braid_big_table
#define x2n_table gdal_x2n_table
#define extra_lbits gdal_extra_lbits
#define extra_dbits gdal_extra_dbits
#define extra_blbits gdal_extra_blbits
#define bl_order gdal_bl_order
#define static_ltree gdal_static_ltree
#define static_dtree gdal_static_dtree
#define base_length gdal_base_length
#define base_dist gdal_base_dist
#define deflate_copyright gdal_deflate_copyright
#define inflate_copyright gdal_inflate_copyright
#define _length_code gdal__length_code
#define _dist_code gdal__dist_code
#define configuration_table gdal_configuration_table
#define static_l_desc gdal_static_l_desc
#define static_d_desc gdal_static_d_desc
#define static_bl_desc gdal_static_bl_desc
#define z_errmsg gdal_z_errmsg
