/* This is a generated file by dump_symbols.h. *DO NOT EDIT MANUALLY !* */
#ifndef GDAL_LIBTIFF_SYMBOL_RENAME_H_INCLUDED
#define GDAL_LIBTIFF_SYMBOL_RENAME_H_INCLUDED
#define add_ms gdal_add_ms
#define allocChoppedUpStripArrays gdal_allocChoppedUpStripArrays
#define buildMap gdal_buildMap
#define BuildMapBitdepth16To8 gdal_BuildMapBitdepth16To8
#define BuildMapUaToAa gdal_BuildMapUaToAa
#define ByteCountLooksBad gdal_ByteCountLooksBad
#define CalcFinalIFDdatasizeReading gdal_CalcFinalIFDdatasizeReading
#define checkcmap gdal_checkcmap
#define CheckDirCount gdal_CheckDirCount
#define ChopUpSingleUncompressedStrip gdal_ChopUpSingleUncompressedStrip
#define CLAMPw gdal_CLAMPw
#define cl_hash gdal_cl_hash
#define cmpTIFFEntryOffsetAndLength gdal_cmpTIFFEntryOffsetAndLength
#define countInkNamesString gdal_countInkNamesString
#define cvtcmap gdal_cvtcmap
#define DoubleToRational gdal_DoubleToRational
#define DoubleToSrational gdal_DoubleToSrational
#define DumpFixupTags gdal_DumpFixupTags
#define DumpModeDecode gdal_DumpModeDecode
#define DumpModeEncode gdal_DumpModeEncode
#define DumpModeSeek gdal_DumpModeSeek
#define equalFuncNumberToOffset gdal_equalFuncNumberToOffset
#define equalFuncOffsetToNumber gdal_equalFuncOffsetToNumber
#define EstimateStripByteCounts gdal_EstimateStripByteCounts
#define EvaluateIFDdatasizeReading gdal_EvaluateIFDdatasizeReading
#define EvaluateIFDdatasizeWrite gdal_EvaluateIFDdatasizeWrite
#define Fax3BadLength gdal_Fax3BadLength
#define Fax3Cleanup gdal_Fax3Cleanup
#define _Fax3Close gdal__Fax3Close
#define Fax3Close gdal_Fax3Close
#define Fax3Decode1D gdal_Fax3Decode1D
#define Fax3Decode2D gdal_Fax3Decode2D
#define Fax3DecodeRLE gdal_Fax3DecodeRLE
#define Fax3Encode gdal_Fax3Encode
#define Fax3Encode1DRow gdal_Fax3Encode1DRow
#define Fax3Encode2DRow gdal_Fax3Encode2DRow
#define Fax3Extension gdal_Fax3Extension
#define Fax3FixupTags gdal_Fax3FixupTags
#define Fax3PostEncode gdal_Fax3PostEncode
#define Fax3PreDecode gdal_Fax3PreDecode
#define Fax3PreEncode gdal_Fax3PreEncode
#define Fax3PrematureEOF gdal_Fax3PrematureEOF
#define Fax3PrintDir gdal_Fax3PrintDir
#define Fax3PutBits gdal_Fax3PutBits
#define Fax3PutEOL gdal_Fax3PutEOL
#define Fax3SetupState gdal_Fax3SetupState
#define Fax3Unexpected gdal_Fax3Unexpected
#define Fax3VGetField gdal_Fax3VGetField
#define Fax3VSetField gdal_Fax3VSetField
#define Fax4Decode gdal_Fax4Decode
#define Fax4Encode gdal_Fax4Encode
#define Fax4PostEncode gdal_Fax4PostEncode
#define find0span gdal_find0span
#define find1span gdal_find1span
#define fpAcc gdal_fpAcc
#define fpDiff gdal_fpDiff
#define frame_dummy gdal_frame_dummy
#define GetLercDataType gdal_GetLercDataType
#define gtStripContig gdal_gtStripContig
#define gtStripSeparate gdal_gtStripSeparate
#define gtTileContig gdal_gtTileContig
#define gtTileSeparate gdal_gtTileSeparate
#define hashFuncNumberToOffset gdal_hashFuncNumberToOffset
#define hashFuncOffsetToNumber gdal_hashFuncOffsetToNumber
#define horAcc16 gdal_horAcc16
#define horAcc32 gdal_horAcc32
#define horAcc64 gdal_horAcc64
#define horAcc8 gdal_horAcc8
#define horDiff16 gdal_horDiff16
#define horDiff32 gdal_horDiff32
#define horDiff64 gdal_horDiff64
#define horDiff8 gdal_horDiff8
#define horizontalAccumulate11 gdal_horizontalAccumulate11
#define horizontalAccumulate12 gdal_horizontalAccumulate12
#define horizontalAccumulate16 gdal_horizontalAccumulate16
#define horizontalAccumulate8 gdal_horizontalAccumulate8
#define horizontalAccumulate8abgr gdal_horizontalAccumulate8abgr
#define horizontalAccumulateF gdal_horizontalAccumulateF
#define horizontalDifference16 gdal_horizontalDifference16
#define horizontalDifference8 gdal_horizontalDifference8
#define horizontalDifferenceF gdal_horizontalDifferenceF
#define InitCCITTFax3 gdal_InitCCITTFax3
#define initCIELabConversion gdal_initCIELabConversion
#define initYCbCrConversion gdal_initYCbCrConversion
#define isCCITTCompression gdal_isCCITTCompression
#define isInRefBlackWhiteRange gdal_isInRefBlackWhiteRange
#define jpeg_create_decompress_encap gdal_jpeg_create_decompress_encap
#define jpeg_encap_unwind gdal_jpeg_encap_unwind
#define jpeg_read_header_encap gdal_jpeg_read_header_encap
#define jpeg_read_raw_data_encap gdal_jpeg_read_raw_data_encap
#define jpeg_read_scanlines_encap gdal_jpeg_read_scanlines_encap
#define jpeg_start_decompress_encap gdal_jpeg_start_decompress_encap
#define L16fromY gdal_L16fromY
#define L16toGry gdal_L16toGry
#define L16toY gdal_L16toY
#define LERCCleanup gdal_LERCCleanup
#define LERCDecode gdal_LERCDecode
#define LERCEncode gdal_LERCEncode
#define LERCFixupTags gdal_LERCFixupTags
#define LERCPostEncode gdal_LERCPostEncode
#define LERCPreDecode gdal_LERCPreDecode
#define LERCPreEncode gdal_LERCPreEncode
#define LERCSetupDecode gdal_LERCSetupDecode
#define LERCSetupEncode gdal_LERCSetupEncode
#define LERCVGetField gdal_LERCVGetField
#define LERCVSetField gdal_LERCVSetField
#define LERCVSetFieldBase gdal_LERCVSetFieldBase
#define LogL10fromY gdal_LogL10fromY
#define LogL10toY gdal_LogL10toY
#define LogL16Decode gdal_LogL16Decode
#define LogL16Encode gdal_LogL16Encode
#define LogL16fromY gdal_LogL16fromY
#define LogL16GuessDataFmt gdal_LogL16GuessDataFmt
#define LogL16InitState gdal_LogL16InitState
#define LogL16toY gdal_LogL16toY
#define LogLuv24fromXYZ gdal_LogLuv24fromXYZ
#define LogLuv24toXYZ gdal_LogLuv24toXYZ
#define LogLuv32fromXYZ gdal_LogLuv32fromXYZ
#define LogLuv32toXYZ gdal_LogLuv32toXYZ
#define LogLuvCleanup gdal_LogLuvCleanup
#define LogLuvClose gdal_LogLuvClose
#define LogLuvDecode24 gdal_LogLuvDecode24
#define LogLuvDecode32 gdal_LogLuvDecode32
#define LogLuvDecodeStrip gdal_LogLuvDecodeStrip
#define LogLuvDecodeTile gdal_LogLuvDecodeTile
#define LogLuvEncode24 gdal_LogLuvEncode24
#define LogLuvEncode32 gdal_LogLuvEncode32
#define LogLuvEncodeStrip gdal_LogLuvEncodeStrip
#define LogLuvEncodeTile gdal_LogLuvEncodeTile
#define LogLuvFixupTags gdal_LogLuvFixupTags
#define LogLuvGuessDataFmt gdal_LogLuvGuessDataFmt
#define LogLuvInitState gdal_LogLuvInitState
#define _logLuvNop gdal__logLuvNop
#define LogLuvSetupDecode gdal_LogLuvSetupDecode
#define LogLuvSetupEncode gdal_LogLuvSetupEncode
#define LogLuvVGetField gdal_LogLuvVGetField
#define LogLuvVSetField gdal_LogLuvVSetField
#define Luv24fromLuv48 gdal_Luv24fromLuv48
#define Luv24fromXYZ gdal_Luv24fromXYZ
#define Luv24toLuv48 gdal_Luv24toLuv48
#define Luv24toRGB gdal_Luv24toRGB
#define Luv24toXYZ gdal_Luv24toXYZ
#define Luv32fromLuv48 gdal_Luv32fromLuv48
#define Luv32fromXYZ gdal_Luv32fromXYZ
#define Luv32toLuv48 gdal_Luv32toLuv48
#define Luv32toRGB gdal_Luv32toRGB
#define Luv32toXYZ gdal_Luv32toXYZ
#define LZMACleanup gdal_LZMACleanup
#define LZMADecode gdal_LZMADecode
#define LZMAEncode gdal_LZMAEncode
#define LZMAFixupTags gdal_LZMAFixupTags
#define LZMAPostEncode gdal_LZMAPostEncode
#define LZMAPreDecode gdal_LZMAPreDecode
#define LZMAPreEncode gdal_LZMAPreEncode
#define LZMASetupDecode gdal_LZMASetupDecode
#define LZMASetupEncode gdal_LZMASetupEncode
#define LZMAStrerror gdal_LZMAStrerror
#define LZMAVGetField gdal_LZMAVGetField
#define LZMAVSetField gdal_LZMAVSetField
#define LZWCleanup gdal_LZWCleanup
#define LZWDecode gdal_LZWDecode
#define LZWDecodeCompat gdal_LZWDecodeCompat
#define LZWEncode gdal_LZWEncode
#define LZWFixupTags gdal_LZWFixupTags
#define LZWPostEncode gdal_LZWPostEncode
#define LZWPreDecode gdal_LZWPreDecode
#define LZWPreEncode gdal_LZWPreEncode
#define LZWSetupDecode gdal_LZWSetupDecode
#define LZWSetupEncode gdal_LZWSetupEncode
#define makebwmap gdal_makebwmap
#define makecmap gdal_makecmap
#define MissingRequired gdal_MissingRequired
#define multiply_ms gdal_multiply_ms
#define multiply_ms gdal_multiply_ms
#define NeXTDecode gdal_NeXTDecode
#define NeXTPreDecode gdal_NeXTPreDecode
#define NoSanitizeSubUInt64 gdal_NoSanitizeSubUInt64
#define _notConfigured gdal__notConfigured
#define NotConfigured gdal_NotConfigured
#define OJPEGCleanup gdal_OJPEGCleanup
#define OJPEGDecode gdal_OJPEGDecode
#define OJPEGDecodeRaw gdal_OJPEGDecodeRaw
#define OJPEGDecodeScanlines gdal_OJPEGDecodeScanlines
#define OJPEGEncode gdal_OJPEGEncode
#define OJPEGFixupTags gdal_OJPEGFixupTags
#define OJPEGLibjpegJpegErrorMgrErrorExit gdal_OJPEGLibjpegJpegErrorMgrErrorExit
#define OJPEGLibjpegJpegErrorMgrOutputMessage gdal_OJPEGLibjpegJpegErrorMgrOutputMessage
#define OJPEGLibjpegJpegSourceMgrFillInputBuffer gdal_OJPEGLibjpegJpegSourceMgrFillInputBuffer
#define OJPEGLibjpegJpegSourceMgrInitSource gdal_OJPEGLibjpegJpegSourceMgrInitSource
#define OJPEGLibjpegJpegSourceMgrResyncToRestart gdal_OJPEGLibjpegJpegSourceMgrResyncToRestart
#define OJPEGLibjpegJpegSourceMgrSkipInputData gdal_OJPEGLibjpegJpegSourceMgrSkipInputData
#define OJPEGLibjpegJpegSourceMgrTermSource gdal_OJPEGLibjpegJpegSourceMgrTermSource
#define OJPEGLibjpegSessionAbort gdal_OJPEGLibjpegSessionAbort
#define OJPEGPostDecode gdal_OJPEGPostDecode
#define OJPEGPostEncode gdal_OJPEGPostEncode
#define OJPEGPreDecode gdal_OJPEGPreDecode
#define OJPEGPreDecodeSkipRaw gdal_OJPEGPreDecodeSkipRaw
#define OJPEGPreDecodeSkipScanlines gdal_OJPEGPreDecodeSkipScanlines
#define OJPEGPreEncode gdal_OJPEGPreEncode
#define OJPEGPrintDir gdal_OJPEGPrintDir
#define OJPEGReadBlock gdal_OJPEGReadBlock
#define OJPEGReadBufferFill gdal_OJPEGReadBufferFill
#define OJPEGReadByte gdal_OJPEGReadByte
#define OJPEGReadByteAdvance gdal_OJPEGReadByteAdvance
#define OJPEGReadBytePeek gdal_OJPEGReadBytePeek
#define OJPEGReadHeaderInfo gdal_OJPEGReadHeaderInfo
#define OJPEGReadHeaderInfoSec gdal_OJPEGReadHeaderInfoSec
#define OJPEGReadHeaderInfoSecStreamDht gdal_OJPEGReadHeaderInfoSecStreamDht
#define OJPEGReadHeaderInfoSecStreamDqt gdal_OJPEGReadHeaderInfoSecStreamDqt
#define OJPEGReadHeaderInfoSecStreamDri gdal_OJPEGReadHeaderInfoSecStreamDri
#define OJPEGReadHeaderInfoSecStreamSof gdal_OJPEGReadHeaderInfoSecStreamSof
#define OJPEGReadHeaderInfoSecStreamSos gdal_OJPEGReadHeaderInfoSecStreamSos
#define OJPEGReadHeaderInfoSecTablesAcTable gdal_OJPEGReadHeaderInfoSecTablesAcTable
#define OJPEGReadHeaderInfoSecTablesDcTable gdal_OJPEGReadHeaderInfoSecTablesDcTable
#define OJPEGReadHeaderInfoSecTablesQTable gdal_OJPEGReadHeaderInfoSecTablesQTable
#define OJPEGReadSecondarySos gdal_OJPEGReadSecondarySos
#define OJPEGReadSkip gdal_OJPEGReadSkip
#define OJPEGReadWord gdal_OJPEGReadWord
#define OJPEGSetupDecode gdal_OJPEGSetupDecode
#define OJPEGSetupEncode gdal_OJPEGSetupEncode
#define OJPEGSubsamplingCorrect gdal_OJPEGSubsamplingCorrect
#define OJPEGVGetField gdal_OJPEGVGetField
#define OJPEGVSetField gdal_OJPEGVSetField
#define OJPEGWriteHeaderInfo gdal_OJPEGWriteHeaderInfo
#define OJPEGWriteStream gdal_OJPEGWriteStream
#define OJPEGWriteStreamAcTable gdal_OJPEGWriteStreamAcTable
#define OJPEGWriteStreamCompressed gdal_OJPEGWriteStreamCompressed
#define OJPEGWriteStreamDcTable gdal_OJPEGWriteStreamDcTable
#define OJPEGWriteStreamDri gdal_OJPEGWriteStreamDri
#define OJPEGWriteStreamEoi gdal_OJPEGWriteStreamEoi
#define OJPEGWriteStreamQTable gdal_OJPEGWriteStreamQTable
#define OJPEGWriteStreamRst gdal_OJPEGWriteStreamRst
#define OJPEGWriteStreamSof gdal_OJPEGWriteStreamSof
#define OJPEGWriteStreamSoi gdal_OJPEGWriteStreamSoi
#define OJPEGWriteStreamSos gdal_OJPEGWriteStreamSos
#define OkToChangeTag gdal_OkToChangeTag
#define oog_encode gdal_oog_encode
#define PackBitsDecode gdal_PackBitsDecode
#define PackBitsEncode gdal_PackBitsEncode
#define PackBitsEncodeChunk gdal_PackBitsEncodeChunk
#define PackBitsPostEncode gdal_PackBitsPostEncode
#define PackBitsPreEncode gdal_PackBitsPreEncode
#define PickContigCase gdal_PickContigCase
#define PickSeparateCase gdal_PickSeparateCase
#define PixarLogCleanup gdal_PixarLogCleanup
#define PixarLogClose gdal_PixarLogClose
#define PixarLogDecode gdal_PixarLogDecode
#define PixarLogEncode gdal_PixarLogEncode
#define PixarLogFixupTags gdal_PixarLogFixupTags
#define PixarLogGuessDataFmt gdal_PixarLogGuessDataFmt
#define PixarLogMakeTables gdal_PixarLogMakeTables
#define PixarLogPostEncode gdal_PixarLogPostEncode
#define PixarLogPreDecode gdal_PixarLogPreDecode
#define PixarLogPreEncode gdal_PixarLogPreEncode
#define PixarLogSetupDecode gdal_PixarLogSetupDecode
#define PixarLogSetupEncode gdal_PixarLogSetupEncode
#define PixarLogVGetField gdal_PixarLogVGetField
#define PixarLogVSetField gdal_PixarLogVSetField
#define PredictorDecodeRow gdal_PredictorDecodeRow
#define PredictorDecodeTile gdal_PredictorDecodeTile
#define PredictorEncodeRow gdal_PredictorEncodeRow
#define PredictorEncodeTile gdal_PredictorEncodeTile
#define PredictorPrintDir gdal_PredictorPrintDir
#define PredictorSetup gdal_PredictorSetup
#define PredictorSetupDecode gdal_PredictorSetupDecode
#define PredictorSetupEncode gdal_PredictorSetupEncode
#define PredictorVGetField gdal_PredictorVGetField
#define PredictorVSetField gdal_PredictorVSetField
#define put16bitbwtile gdal_put16bitbwtile
#define put1bitbwtile gdal_put1bitbwtile
#define put1bitcmaptile gdal_put1bitcmaptile
#define put2bitbwtile gdal_put2bitbwtile
#define put2bitcmaptile gdal_put2bitcmaptile
#define put4bitbwtile gdal_put4bitbwtile
#define put4bitcmaptile gdal_put4bitcmaptile
#define put8bitcmaptile gdal_put8bitcmaptile
#define putagreytile gdal_putagreytile
#define putCMYKseparate8bittile gdal_putCMYKseparate8bittile
#define putcontig8bitCIELab16 gdal_putcontig8bitCIELab16
#define putcontig8bitCIELab8 gdal_putcontig8bitCIELab8
#define putcontig8bitYCbCr11tile gdal_putcontig8bitYCbCr11tile
#define putcontig8bitYCbCr12tile gdal_putcontig8bitYCbCr12tile
#define putcontig8bitYCbCr21tile gdal_putcontig8bitYCbCr21tile
#define putcontig8bitYCbCr22tile gdal_putcontig8bitYCbCr22tile
#define putcontig8bitYCbCr41tile gdal_putcontig8bitYCbCr41tile
#define putcontig8bitYCbCr42tile gdal_putcontig8bitYCbCr42tile
#define putcontig8bitYCbCr44tile gdal_putcontig8bitYCbCr44tile
#define putgreytile gdal_putgreytile
#define putRGBAAcontig16bittile gdal_putRGBAAcontig16bittile
#define putRGBAAcontig8bittile gdal_putRGBAAcontig8bittile
#define putRGBAAseparate16bittile gdal_putRGBAAseparate16bittile
#define putRGBAAseparate8bittile gdal_putRGBAAseparate8bittile
#define putRGBcontig16bittile gdal_putRGBcontig16bittile
#define putRGBcontig8bitCMYKMaptile gdal_putRGBcontig8bitCMYKMaptile
#define putRGBcontig8bitCMYKtile gdal_putRGBcontig8bitCMYKtile
#define putRGBcontig8bittile gdal_putRGBcontig8bittile
#define putRGBseparate16bittile gdal_putRGBseparate16bittile
#define putRGBseparate8bittile gdal_putRGBseparate8bittile
#define putRGBUAcontig16bittile gdal_putRGBUAcontig16bittile
#define putRGBUAcontig8bittile gdal_putRGBUAcontig8bittile
#define putRGBUAseparate16bittile gdal_putRGBUAseparate16bittile
#define putRGBUAseparate8bittile gdal_putRGBUAseparate8bittile
#define putseparate8bitYCbCr11tile gdal_putseparate8bitYCbCr11tile
#define putspan gdal_putspan
#define setByteArray gdal_setByteArray
#define setDoubleArrayOneValue gdal_setDoubleArrayOneValue
#define setExtraSamples gdal_setExtraSamples
#define setorientation gdal_setorientation
#define SetupBuffers gdal_SetupBuffers
#define setupMap gdal_setupMap
#define swabHorAcc16 gdal_swabHorAcc16
#define swabHorAcc32 gdal_swabHorAcc32
#define swabHorAcc64 gdal_swabHorAcc64
#define swabHorDiff16 gdal_swabHorDiff16
#define swabHorDiff32 gdal_swabHorDiff32
#define swabHorDiff64 gdal_swabHorDiff64
#define tagCompare gdal_tagCompare
#define tagNameCompare gdal_tagNameCompare
#define td_lfind gdal_td_lfind
#define ThunderDecode gdal_ThunderDecode
#define ThunderDecodeRow gdal_ThunderDecodeRow
#define ThunderSetupDecode gdal_ThunderSetupDecode
#define TIFFAccessTagMethods gdal_TIFFAccessTagMethods
#define TIFFAdvanceDirectory gdal_TIFFAdvanceDirectory
#define TIFFAppendToStrip gdal_TIFFAppendToStrip
#define _TIFFcalloc gdal__TIFFcalloc
#define _TIFFcallocExt gdal__TIFFcallocExt
#define _TIFFCastUInt64ToSSize gdal__TIFFCastUInt64ToSSize
#define _TIFFCheckDirNumberAndOffset gdal__TIFFCheckDirNumberAndOffset
#define _TIFFCheckFieldIsValidForCodec gdal__TIFFCheckFieldIsValidForCodec
#define _TIFFCheckMalloc gdal__TIFFCheckMalloc
#define TIFFCheckpointDirectory gdal_TIFFCheckpointDirectory
#define TIFFCheckRead gdal_TIFFCheckRead
#define _TIFFCheckRealloc gdal__TIFFCheckRealloc
#define TIFFCheckTile gdal_TIFFCheckTile
#define TIFFCIELab16ToXYZ gdal_TIFFCIELab16ToXYZ
#define TIFFCIELabToRGBInit gdal_TIFFCIELabToRGBInit
#define TIFFCIELabToXYZ gdal_TIFFCIELabToXYZ
#define _TIFFClampDoubleToFloat gdal__TIFFClampDoubleToFloat
#define TIFFClampDoubleToInt16 gdal_TIFFClampDoubleToInt16
#define TIFFClampDoubleToInt32 gdal_TIFFClampDoubleToInt32
#define TIFFClampDoubleToInt8 gdal_TIFFClampDoubleToInt8
#define TIFFClampDoubleToUInt16 gdal_TIFFClampDoubleToUInt16
#define _TIFFClampDoubleToUInt32 gdal__TIFFClampDoubleToUInt32
#define TIFFClampDoubleToUInt32 gdal_TIFFClampDoubleToUInt32
#define TIFFClampDoubleToUInt8 gdal_TIFFClampDoubleToUInt8
#define TIFFCleanup gdal_TIFFCleanup
#define _TIFFCleanupIFDOffsetAndNumberMaps gdal__TIFFCleanupIFDOffsetAndNumberMaps
#define TIFFClientdata gdal_TIFFClientdata
#define TIFFClientOpen gdal_TIFFClientOpen
#define TIFFClientOpenExt gdal_TIFFClientOpenExt
#define TIFFClose gdal_TIFFClose
#define _tiffCloseProc gdal__tiffCloseProc
#define TIFFComputeStrip gdal_TIFFComputeStrip
#define TIFFComputeTile gdal_TIFFComputeTile
#define _TIFFCreateAnonField gdal__TIFFCreateAnonField
#define TIFFCreateCustomDirectory gdal_TIFFCreateCustomDirectory
#define TIFFCreateDirectory gdal_TIFFCreateDirectory
#define TIFFCreateEXIFDirectory gdal_TIFFCreateEXIFDirectory
#define TIFFCreateGPSDirectory gdal_TIFFCreateGPSDirectory
#define TIFFCurrentDirectory gdal_TIFFCurrentDirectory
#define TIFFCurrentDirOffset gdal_TIFFCurrentDirOffset
#define TIFFCurrentRow gdal_TIFFCurrentRow
#define TIFFCurrentStrip gdal_TIFFCurrentStrip
#define TIFFCurrentTile gdal_TIFFCurrentTile
#define TIFFDataWidth gdal_TIFFDataWidth
#define TIFFDefaultDirectory gdal_TIFFDefaultDirectory
#define TIFFDefaultRefBlackWhite gdal_TIFFDefaultRefBlackWhite
#define _TIFFDefaultStripSize gdal__TIFFDefaultStripSize
#define TIFFDefaultStripSize gdal_TIFFDefaultStripSize
#define _TIFFDefaultTileSize gdal__TIFFDefaultTileSize
#define TIFFDefaultTileSize gdal_TIFFDefaultTileSize
#define TIFFDefaultTransferFunction gdal_TIFFDefaultTransferFunction
#define TIFFDeferStrileArrayWriting gdal_TIFFDeferStrileArrayWriting
#define _tiffDummyMapProc gdal__tiffDummyMapProc
#define _tiffDummyUnmapProc gdal__tiffDummyUnmapProc
#define _TIFFEmitErrorAboveMaxCumulatedMemAlloc gdal__TIFFEmitErrorAboveMaxCumulatedMemAlloc
#define _TIFFEmitErrorAboveMaxSingleMemAlloc gdal__TIFFEmitErrorAboveMaxSingleMemAlloc
#define TIFFError gdal_TIFFError
#define _TIFFErrorEarly gdal__TIFFErrorEarly
#define TIFFErrorExt gdal_TIFFErrorExt
#define TIFFErrorExtR gdal_TIFFErrorExtR
#define _TIFFFax3fillruns gdal__TIFFFax3fillruns
#define TIFFFdOpen gdal_TIFFFdOpen
#define TIFFFetchDirectory gdal_TIFFFetchDirectory
#define TIFFFetchNormalTag gdal_TIFFFetchNormalTag
#define _TIFFFetchStrileValue gdal__TIFFFetchStrileValue
#define TIFFFetchStripThing gdal_TIFFFetchStripThing
#define TIFFFetchSubjectDistance gdal_TIFFFetchSubjectDistance
#define TIFFFieldDataType gdal_TIFFFieldDataType
#define TIFFFieldIsAnonymous gdal_TIFFFieldIsAnonymous
#define TIFFFieldName gdal_TIFFFieldName
#define TIFFFieldPassCount gdal_TIFFFieldPassCount
#define TIFFFieldReadCount gdal_TIFFFieldReadCount
#define TIFFFieldSetGetCountSize gdal_TIFFFieldSetGetCountSize
#define TIFFFieldSetGetSize gdal_TIFFFieldSetGetSize
#define TIFFFieldTag gdal_TIFFFieldTag
#define TIFFFieldWithName gdal_TIFFFieldWithName
#define TIFFFieldWithTag gdal_TIFFFieldWithTag
#define TIFFFieldWriteCount gdal_TIFFFieldWriteCount
#define TIFFFileName gdal_TIFFFileName
#define TIFFFileno gdal_TIFFFileno
#define _TIFFFillStriles gdal__TIFFFillStriles
#define _TIFFFillStrilesInternal gdal__TIFFFillStrilesInternal
#define TIFFFillStrip gdal_TIFFFillStrip
#define TIFFFillStripPartial gdal_TIFFFillStripPartial
#define TIFFFillTile gdal_TIFFFillTile
#define TIFFFindCODEC gdal_TIFFFindCODEC
#define TIFFFindField gdal_TIFFFindField
#define _TIFFFindFieldByName gdal__TIFFFindFieldByName
#define _TIFFFindOrRegisterField gdal__TIFFFindOrRegisterField
#define TIFFFlush gdal_TIFFFlush
#define TIFFFlushData gdal_TIFFFlushData
#define TIFFFlushData1 gdal_TIFFFlushData1
#define TIFFForceStrileArrayWriting gdal_TIFFForceStrileArrayWriting
#define _TIFFfree gdal__TIFFfree
#define TIFFFreeDirectory gdal_TIFFFreeDirectory
#define _TIFFfreeExt gdal__TIFFfreeExt
#define TIFFGetBitRevTable gdal_TIFFGetBitRevTable
#define TIFFGetClientInfo gdal_TIFFGetClientInfo
#define TIFFGetCloseProc gdal_TIFFGetCloseProc
#define TIFFGetConfiguredCODECs gdal_TIFFGetConfiguredCODECs
#define _TIFFGetDirNumberFromOffset gdal__TIFFGetDirNumberFromOffset
#define _TIFFGetExifFields gdal__TIFFGetExifFields
#define TIFFGetField gdal_TIFFGetField
#define TIFFGetFieldDefaulted gdal_TIFFGetFieldDefaulted
#define _TIFFGetFields gdal__TIFFGetFields
#define _TIFFGetGpsFields gdal__TIFFGetGpsFields
#define TIFFGetMapFileProc gdal_TIFFGetMapFileProc
#define _TIFFGetMaxColorChannels gdal__TIFFGetMaxColorChannels
#define _TIFFgetMode gdal__TIFFgetMode
#define TIFFGetMode gdal_TIFFGetMode
#define _TIFFGetOffsetFromDirNumber gdal__TIFFGetOffsetFromDirNumber
#define TIFFGetReadProc gdal_TIFFGetReadProc
#define TIFFGetSeekProc gdal_TIFFGetSeekProc
#define TIFFGetSizeProc gdal_TIFFGetSizeProc
#define TIFFGetStrileByteCount gdal_TIFFGetStrileByteCount
#define TIFFGetStrileByteCountWithErr gdal_TIFFGetStrileByteCountWithErr
#define TIFFGetStrileOffset gdal_TIFFGetStrileOffset
#define _TIFFGetStrileOffsetOrByteCountValue gdal__TIFFGetStrileOffsetOrByteCountValue
#define TIFFGetStrileOffsetWithErr gdal_TIFFGetStrileOffsetWithErr
#define TIFFGetTagListCount gdal_TIFFGetTagListCount
#define TIFFGetTagListEntry gdal_TIFFGetTagListEntry
#define TIFFGetUnmapFileProc gdal_TIFFGetUnmapFileProc
#define TIFFGetVersion gdal_TIFFGetVersion
#define TIFFGetWriteProc gdal_TIFFGetWriteProc
#define TIFFGrowStrips gdal_TIFFGrowStrips
#define TIFFHashSetClearInternal gdal_TIFFHashSetClearInternal
#define TIFFHashSetDestroy gdal_TIFFHashSetDestroy
#define TIFFHashSetEqualPointer gdal_TIFFHashSetEqualPointer
#define TIFFHashSetFindPtr gdal_TIFFHashSetFindPtr
#define TIFFHashSetGetNewListElt gdal_TIFFHashSetGetNewListElt
#define TIFFHashSetHashPointer gdal_TIFFHashSetHashPointer
#define TIFFHashSetInsert gdal_TIFFHashSetInsert
#define TIFFHashSetLookup gdal_TIFFHashSetLookup
#define TIFFHashSetNew gdal_TIFFHashSetNew
#define TIFFHashSetRehash gdal_TIFFHashSetRehash
#define TIFFHashSetRemove gdal_TIFFHashSetRemove
#define TIFFHashSetRemoveInternal gdal_TIFFHashSetRemoveInternal
#define TIFFHashSetReturnListElt gdal_TIFFHashSetReturnListElt
#define TIFFHashSetSize gdal_TIFFHashSetSize
#define TIFFIsBigEndian gdal_TIFFIsBigEndian
#define TIFFIsBigTIFF gdal_TIFFIsBigTIFF
#define TIFFIsByteSwapped gdal_TIFFIsByteSwapped
#define TIFFIsCODECConfigured gdal_TIFFIsCODECConfigured
#define TIFFIsMSB2LSB gdal_TIFFIsMSB2LSB
#define TIFFIsTiled gdal_TIFFIsTiled
#define TIFFIsUpSampled gdal_TIFFIsUpSampled
#define tiff_itrunc gdal_tiff_itrunc
#define TIFFLastDirectory gdal_TIFFLastDirectory
#define TIFFLinkDirectory gdal_TIFFLinkDirectory
#define TIFFListDestroy gdal_TIFFListDestroy
#define _TIFFmalloc gdal__TIFFmalloc
#define _TIFFmallocExt gdal__TIFFmallocExt
#define _tiffMapProc gdal__tiffMapProc
#define _TIFFmemcmp gdal__TIFFmemcmp
#define _TIFFmemcpy gdal__TIFFmemcpy
#define _TIFFmemset gdal__TIFFmemset
#define TIFFMergeFieldInfo gdal_TIFFMergeFieldInfo
#define _TIFFMergeFields gdal__TIFFMergeFields
#define _TIFFMultiply32 gdal__TIFFMultiply32
#define _TIFFMultiply64 gdal__TIFFMultiply64
#define _TIFFMultiplySSize gdal__TIFFMultiplySSize
#define TIFFNoDecode gdal_TIFFNoDecode
#define TIFFNoEncode gdal_TIFFNoEncode
#define _TIFFNoFixupTags gdal__TIFFNoFixupTags
#define _TIFFNoPostDecode gdal__TIFFNoPostDecode
#define _TIFFNoPreCode gdal__TIFFNoPreCode
#define _TIFFNoRowDecode gdal__TIFFNoRowDecode
#define _TIFFNoRowEncode gdal__TIFFNoRowEncode
#define _TIFFNoSeek gdal__TIFFNoSeek
#define _TIFFNoStripDecode gdal__TIFFNoStripDecode
#define _TIFFNoStripEncode gdal__TIFFNoStripEncode
#define _TIFFNoTileDecode gdal__TIFFNoTileDecode
#define _TIFFNoTileEncode gdal__TIFFNoTileEncode
#define TIFFNumberOfDirectories gdal_TIFFNumberOfDirectories
#define TIFFNumberOfStrips gdal_TIFFNumberOfStrips
#define TIFFNumberOfTiles gdal_TIFFNumberOfTiles
#define TIFFOpen gdal_TIFFOpen
#define TIFFOpenOptionsAlloc gdal_TIFFOpenOptionsAlloc
#define TIFFOpenOptionsFree gdal_TIFFOpenOptionsFree
#define TIFFOpenOptionsSetErrorHandlerExtR gdal_TIFFOpenOptionsSetErrorHandlerExtR
#define TIFFOpenOptionsSetMaxCumulatedMemAlloc gdal_TIFFOpenOptionsSetMaxCumulatedMemAlloc
#define TIFFOpenOptionsSetMaxSingleMemAlloc gdal_TIFFOpenOptionsSetMaxSingleMemAlloc
#define TIFFOpenOptionsSetWarnAboutUnknownTags gdal_TIFFOpenOptionsSetWarnAboutUnknownTags
#define TIFFOpenOptionsSetWarningHandlerExtR gdal_TIFFOpenOptionsSetWarningHandlerExtR
#define _TIFFPartialReadStripArray gdal__TIFFPartialReadStripArray
#define TIFFPredictorCleanup gdal_TIFFPredictorCleanup
#define TIFFPredictorInit gdal_TIFFPredictorInit
#define _TIFFPrettyPrintField gdal__TIFFPrettyPrintField
#define _TIFFprintAscii gdal__TIFFprintAscii
#define _TIFFprintAsciiBounded gdal__TIFFprintAsciiBounded
#define _TIFFprintAsciiTag gdal__TIFFprintAsciiTag
#define TIFFPrintDirectory gdal_TIFFPrintDirectory
#define _TIFFPrintField gdal__TIFFPrintField
#define _TIFFPrintFieldInfo gdal__TIFFPrintFieldInfo
#define TIFFRasterScanlineSize gdal_TIFFRasterScanlineSize
#define TIFFRasterScanlineSize64 gdal_TIFFRasterScanlineSize64
#define TIFFRawStripSize gdal_TIFFRawStripSize
#define TIFFRawStripSize64 gdal_TIFFRawStripSize64
#define TIFFReadAndRealloc gdal_TIFFReadAndRealloc
#define TIFFReadBufferSetup gdal_TIFFReadBufferSetup
#define TIFFReadCustomDirectory gdal_TIFFReadCustomDirectory
#define TIFFReadDirectory gdal_TIFFReadDirectory
#define TIFFReadDirectoryCheckOrder gdal_TIFFReadDirectoryCheckOrder
#define TIFFReadDirectoryFindEntry gdal_TIFFReadDirectoryFindEntry
#define TIFFReadDirectoryFindFieldInfo gdal_TIFFReadDirectoryFindFieldInfo
#define TIFFReadDirEntryArray gdal_TIFFReadDirEntryArray
#define TIFFReadDirEntryArrayWithLimit gdal_TIFFReadDirEntryArrayWithLimit
#define TIFFReadDirEntryByte gdal_TIFFReadDirEntryByte
#define TIFFReadDirEntryByteArray gdal_TIFFReadDirEntryByteArray
#define TIFFReadDirEntryCheckedByte gdal_TIFFReadDirEntryCheckedByte
#define TIFFReadDirEntryCheckedDouble gdal_TIFFReadDirEntryCheckedDouble
#define TIFFReadDirEntryCheckedFloat gdal_TIFFReadDirEntryCheckedFloat
#define TIFFReadDirEntryCheckedLong gdal_TIFFReadDirEntryCheckedLong
#define TIFFReadDirEntryCheckedLong8 gdal_TIFFReadDirEntryCheckedLong8
#define TIFFReadDirEntryCheckedRational gdal_TIFFReadDirEntryCheckedRational
#define TIFFReadDirEntryCheckedSbyte gdal_TIFFReadDirEntryCheckedSbyte
#define TIFFReadDirEntryCheckedShort gdal_TIFFReadDirEntryCheckedShort
#define TIFFReadDirEntryCheckedSlong gdal_TIFFReadDirEntryCheckedSlong
#define TIFFReadDirEntryCheckedSlong8 gdal_TIFFReadDirEntryCheckedSlong8
#define TIFFReadDirEntryCheckedSrational gdal_TIFFReadDirEntryCheckedSrational
#define TIFFReadDirEntryCheckedSshort gdal_TIFFReadDirEntryCheckedSshort
#define TIFFReadDirEntryCheckRangeByteLong gdal_TIFFReadDirEntryCheckRangeByteLong
#define TIFFReadDirEntryCheckRangeByteLong8 gdal_TIFFReadDirEntryCheckRangeByteLong8
#define TIFFReadDirEntryCheckRangeByteSbyte gdal_TIFFReadDirEntryCheckRangeByteSbyte
#define TIFFReadDirEntryCheckRangeByteShort gdal_TIFFReadDirEntryCheckRangeByteShort
#define TIFFReadDirEntryCheckRangeByteSlong gdal_TIFFReadDirEntryCheckRangeByteSlong
#define TIFFReadDirEntryCheckRangeByteSlong8 gdal_TIFFReadDirEntryCheckRangeByteSlong8
#define TIFFReadDirEntryCheckRangeByteSshort gdal_TIFFReadDirEntryCheckRangeByteSshort
#define TIFFReadDirEntryCheckRangeLong8Sbyte gdal_TIFFReadDirEntryCheckRangeLong8Sbyte
#define TIFFReadDirEntryCheckRangeLong8Slong gdal_TIFFReadDirEntryCheckRangeLong8Slong
#define TIFFReadDirEntryCheckRangeLong8Slong8 gdal_TIFFReadDirEntryCheckRangeLong8Slong8
#define TIFFReadDirEntryCheckRangeLong8Sshort gdal_TIFFReadDirEntryCheckRangeLong8Sshort
#define TIFFReadDirEntryCheckRangeLongLong8 gdal_TIFFReadDirEntryCheckRangeLongLong8
#define TIFFReadDirEntryCheckRangeLongSbyte gdal_TIFFReadDirEntryCheckRangeLongSbyte
#define TIFFReadDirEntryCheckRangeLongSlong gdal_TIFFReadDirEntryCheckRangeLongSlong
#define TIFFReadDirEntryCheckRangeLongSlong8 gdal_TIFFReadDirEntryCheckRangeLongSlong8
#define TIFFReadDirEntryCheckRangeLongSshort gdal_TIFFReadDirEntryCheckRangeLongSshort
#define TIFFReadDirEntryCheckRangeSbyteByte gdal_TIFFReadDirEntryCheckRangeSbyteByte
#define TIFFReadDirEntryCheckRangeSbyteLong gdal_TIFFReadDirEntryCheckRangeSbyteLong
#define TIFFReadDirEntryCheckRangeSbyteLong8 gdal_TIFFReadDirEntryCheckRangeSbyteLong8
#define TIFFReadDirEntryCheckRangeSbyteShort gdal_TIFFReadDirEntryCheckRangeSbyteShort
#define TIFFReadDirEntryCheckRangeSbyteSlong gdal_TIFFReadDirEntryCheckRangeSbyteSlong
#define TIFFReadDirEntryCheckRangeSbyteSlong8 gdal_TIFFReadDirEntryCheckRangeSbyteSlong8
#define TIFFReadDirEntryCheckRangeSbyteSshort gdal_TIFFReadDirEntryCheckRangeSbyteSshort
#define TIFFReadDirEntryCheckRangeShortLong gdal_TIFFReadDirEntryCheckRangeShortLong
#define TIFFReadDirEntryCheckRangeShortLong8 gdal_TIFFReadDirEntryCheckRangeShortLong8
#define TIFFReadDirEntryCheckRangeShortSbyte gdal_TIFFReadDirEntryCheckRangeShortSbyte
#define TIFFReadDirEntryCheckRangeShortSlong gdal_TIFFReadDirEntryCheckRangeShortSlong
#define TIFFReadDirEntryCheckRangeShortSlong8 gdal_TIFFReadDirEntryCheckRangeShortSlong8
#define TIFFReadDirEntryCheckRangeShortSshort gdal_TIFFReadDirEntryCheckRangeShortSshort
#define TIFFReadDirEntryCheckRangeSlong8Long8 gdal_TIFFReadDirEntryCheckRangeSlong8Long8
#define TIFFReadDirEntryCheckRangeSlongLong gdal_TIFFReadDirEntryCheckRangeSlongLong
#define TIFFReadDirEntryCheckRangeSlongLong8 gdal_TIFFReadDirEntryCheckRangeSlongLong8
#define TIFFReadDirEntryCheckRangeSlongSlong8 gdal_TIFFReadDirEntryCheckRangeSlongSlong8
#define TIFFReadDirEntryCheckRangeSshortLong gdal_TIFFReadDirEntryCheckRangeSshortLong
#define TIFFReadDirEntryCheckRangeSshortLong8 gdal_TIFFReadDirEntryCheckRangeSshortLong8
#define TIFFReadDirEntryCheckRangeSshortShort gdal_TIFFReadDirEntryCheckRangeSshortShort
#define TIFFReadDirEntryCheckRangeSshortSlong gdal_TIFFReadDirEntryCheckRangeSshortSlong
#define TIFFReadDirEntryCheckRangeSshortSlong8 gdal_TIFFReadDirEntryCheckRangeSshortSlong8
#define TIFFReadDirEntryData gdal_TIFFReadDirEntryData
#define TIFFReadDirEntryDataAndRealloc gdal_TIFFReadDirEntryDataAndRealloc
#define TIFFReadDirEntryDouble gdal_TIFFReadDirEntryDouble
#define TIFFReadDirEntryDoubleArray gdal_TIFFReadDirEntryDoubleArray
#define TIFFReadDirEntryFloat gdal_TIFFReadDirEntryFloat
#define TIFFReadDirEntryFloatArray gdal_TIFFReadDirEntryFloatArray
#define TIFFReadDirEntryIfd8 gdal_TIFFReadDirEntryIfd8
#define TIFFReadDirEntryIfd8Array gdal_TIFFReadDirEntryIfd8Array
#define TIFFReadDirEntryLong gdal_TIFFReadDirEntryLong
#define TIFFReadDirEntryLong8 gdal_TIFFReadDirEntryLong8
#define TIFFReadDirEntryLong8Array gdal_TIFFReadDirEntryLong8Array
#define TIFFReadDirEntryLong8ArrayWithLimit gdal_TIFFReadDirEntryLong8ArrayWithLimit
#define TIFFReadDirEntryLongArray gdal_TIFFReadDirEntryLongArray
#define TIFFReadDirEntryOutputErr gdal_TIFFReadDirEntryOutputErr
#define TIFFReadDirEntryPersampleShort gdal_TIFFReadDirEntryPersampleShort
#define TIFFReadDirEntrySbyte gdal_TIFFReadDirEntrySbyte
#define TIFFReadDirEntrySbyteArray gdal_TIFFReadDirEntrySbyteArray
#define TIFFReadDirEntryShort gdal_TIFFReadDirEntryShort
#define TIFFReadDirEntryShortArray gdal_TIFFReadDirEntryShortArray
#define TIFFReadDirEntrySlong gdal_TIFFReadDirEntrySlong
#define TIFFReadDirEntrySlong8 gdal_TIFFReadDirEntrySlong8
#define TIFFReadDirEntrySlong8Array gdal_TIFFReadDirEntrySlong8Array
#define TIFFReadDirEntrySlongArray gdal_TIFFReadDirEntrySlongArray
#define TIFFReadDirEntrySshort gdal_TIFFReadDirEntrySshort
#define TIFFReadDirEntrySshortArray gdal_TIFFReadDirEntrySshortArray
#define TIFFReadEncodedStrip gdal_TIFFReadEncodedStrip
#define _TIFFReadEncodedStripAndAllocBuffer gdal__TIFFReadEncodedStripAndAllocBuffer
#define TIFFReadEncodedStripGetStripSize gdal_TIFFReadEncodedStripGetStripSize
#define TIFFReadEncodedTile gdal_TIFFReadEncodedTile
#define _TIFFReadEncodedTileAndAllocBuffer gdal__TIFFReadEncodedTileAndAllocBuffer
#define TIFFReadEXIFDirectory gdal_TIFFReadEXIFDirectory
#define TIFFReadFromUserBuffer gdal_TIFFReadFromUserBuffer
#define TIFFReadGPSDirectory gdal_TIFFReadGPSDirectory
#define _tiffReadProc gdal__tiffReadProc
#define TIFFReadRawStrip gdal_TIFFReadRawStrip
#define TIFFReadRawStrip1 gdal_TIFFReadRawStrip1
#define TIFFReadRawStripOrTile2 gdal_TIFFReadRawStripOrTile2
#define TIFFReadRawTile gdal_TIFFReadRawTile
#define TIFFReadRawTile1 gdal_TIFFReadRawTile1
#define TIFFReadRGBAImage gdal_TIFFReadRGBAImage
#define TIFFReadRGBAImageOriented gdal_TIFFReadRGBAImageOriented
#define TIFFReadRGBAStrip gdal_TIFFReadRGBAStrip
#define TIFFReadRGBAStripExt gdal_TIFFReadRGBAStripExt
#define TIFFReadRGBATile gdal_TIFFReadRGBATile
#define TIFFReadRGBATileExt gdal_TIFFReadRGBATileExt
#define TIFFReadScanline gdal_TIFFReadScanline
#define TIFFReadTile gdal_TIFFReadTile
#define _TIFFReadTileAndAllocBuffer gdal__TIFFReadTileAndAllocBuffer
#define TIFFReadUInt64 gdal_TIFFReadUInt64
#define _TIFFrealloc gdal__TIFFrealloc
#define _TIFFreallocExt gdal__TIFFreallocExt
#define TIFFRegisterCODEC gdal_TIFFRegisterCODEC
#define _TIFFRemoveEntryFromDirectoryListByOffset gdal__TIFFRemoveEntryFromDirectoryListByOffset
#define _TIFFReserveLargeEnoughWriteBuffer gdal__TIFFReserveLargeEnoughWriteBuffer
#define TIFFReverseBits gdal_TIFFReverseBits
#define TIFFRewriteDirectory gdal_TIFFRewriteDirectory
#define TIFFRewriteDirectorySec gdal_TIFFRewriteDirectorySec
#define _TIFFRewriteField gdal__TIFFRewriteField
#define TIFFRGBAImageBegin gdal_TIFFRGBAImageBegin
#define TIFFRGBAImageEnd gdal_TIFFRGBAImageEnd
#define TIFFRGBAImageGet gdal_TIFFRGBAImageGet
#define TIFFRGBAImageOK gdal_TIFFRGBAImageOK
#define TIFFScanlineSize gdal_TIFFScanlineSize
#define TIFFScanlineSize64 gdal_TIFFScanlineSize64
#define TIFFSeek gdal_TIFFSeek
#define _TIFFSeekOK gdal__TIFFSeekOK
#define _tiffSeekProc gdal__tiffSeekProc
#define _TIFFsetByteArray gdal__TIFFsetByteArray
#define _TIFFsetByteArrayExt gdal__TIFFsetByteArrayExt
#define TIFFSetClientdata gdal_TIFFSetClientdata
#define TIFFSetClientInfo gdal_TIFFSetClientInfo
#define TIFFSetCompressionScheme gdal_TIFFSetCompressionScheme
#define _TIFFSetDefaultCompressionState gdal__TIFFSetDefaultCompressionState
#define TIFFSetDirectory gdal_TIFFSetDirectory
#define _TIFFsetDoubleArray gdal__TIFFsetDoubleArray
#define _TIFFsetDoubleArrayExt gdal__TIFFsetDoubleArrayExt
#define TIFFSetErrorHandler gdal_TIFFSetErrorHandler
#define TIFFSetErrorHandlerExt gdal_TIFFSetErrorHandlerExt
#define TIFFSetField gdal_TIFFSetField
#define TIFFSetFileName gdal_TIFFSetFileName
#define TIFFSetFileno gdal_TIFFSetFileno
#define _TIFFsetFloatArray gdal__TIFFsetFloatArray
#define _TIFFsetFloatArrayExt gdal__TIFFsetFloatArrayExt
#define _TIFFSetGetType gdal__TIFFSetGetType
#define _TIFFsetLong8Array gdal__TIFFsetLong8Array
#define _TIFFsetLongArray gdal__TIFFsetLongArray
#define _TIFFsetLongArrayExt gdal__TIFFsetLongArrayExt
#define TIFFSetMode gdal_TIFFSetMode
#define _TIFFsetNString gdal__TIFFsetNString
#define _TIFFsetShortArray gdal__TIFFsetShortArray
#define _TIFFsetShortArrayExt gdal__TIFFsetShortArrayExt
#define TIFFSetSubDirectory gdal_TIFFSetSubDirectory
#define TIFFSetTagExtender gdal_TIFFSetTagExtender
#define _TIFFSetupFields gdal__TIFFSetupFields
#define TIFFSetupStrips gdal_TIFFSetupStrips
#define TIFFSetWarningHandler gdal_TIFFSetWarningHandler
#define TIFFSetWarningHandlerExt gdal_TIFFSetWarningHandlerExt
#define TIFFSetWriteOffset gdal_TIFFSetWriteOffset
#define _tiffSizeProc gdal__tiffSizeProc
#define TIFFStartStrip gdal_TIFFStartStrip
#define TIFFStartTile gdal_TIFFStartTile
#define TIFFStripSize gdal_TIFFStripSize
#define TIFFStripSize64 gdal_TIFFStripSize64
#define _TIFFSwab16BitData gdal__TIFFSwab16BitData
#define _TIFFSwab24BitData gdal__TIFFSwab24BitData
#define _TIFFSwab32BitData gdal__TIFFSwab32BitData
#define _TIFFSwab64BitData gdal__TIFFSwab64BitData
#define TIFFSwabArrayOfDouble gdal_TIFFSwabArrayOfDouble
#define TIFFSwabArrayOfFloat gdal_TIFFSwabArrayOfFloat
#define TIFFSwabArrayOfLong gdal_TIFFSwabArrayOfLong
#define TIFFSwabArrayOfLong8 gdal_TIFFSwabArrayOfLong8
#define TIFFSwabArrayOfShort gdal_TIFFSwabArrayOfShort
#define TIFFSwabArrayOfTriples gdal_TIFFSwabArrayOfTriples
#define TIFFSwabDouble gdal_TIFFSwabDouble
#define TIFFSwabFloat gdal_TIFFSwabFloat
#define TIFFSwabLong gdal_TIFFSwabLong
#define TIFFSwabLong8 gdal_TIFFSwabLong8
#define TIFFSwabShort gdal_TIFFSwabShort
#define TIFFTileRowSize gdal_TIFFTileRowSize
#define TIFFTileRowSize64 gdal_TIFFTileRowSize64
#define TIFFTileSize gdal_TIFFTileSize
#define TIFFTileSize64 gdal_TIFFTileSize64
#define _TIFFtrue gdal__TIFFtrue
#define TIFFUnlinkDirectory gdal_TIFFUnlinkDirectory
#define _tiffUnmapProc gdal__tiffUnmapProc
#define TIFFUnRegisterCODEC gdal_TIFFUnRegisterCODEC
#define _TIFFUnsanitizedAddUInt64AndInt gdal__TIFFUnsanitizedAddUInt64AndInt
#define TIFFUnsetField gdal_TIFFUnsetField
#define _TIFFVGetField gdal__TIFFVGetField
#define TIFFVGetField gdal_TIFFVGetField
#define TIFFVGetFieldDefaulted gdal_TIFFVGetFieldDefaulted
#define _TIFFvoid gdal__TIFFvoid
#define _TIFFVSetField gdal__TIFFVSetField
#define TIFFVSetField gdal_TIFFVSetField
#define TIFFVStripSize gdal_TIFFVStripSize
#define TIFFVStripSize64 gdal_TIFFVStripSize64
#define TIFFVTileSize gdal_TIFFVTileSize
#define TIFFVTileSize64 gdal_TIFFVTileSize64
#define TIFFWarning gdal_TIFFWarning
#define TIFFWarningExt gdal_TIFFWarningExt
#define TIFFWarningExtR gdal_TIFFWarningExtR
#define TIFFWriteBufferSetup gdal_TIFFWriteBufferSetup
#define TIFFWriteCheck gdal_TIFFWriteCheck
#define TIFFWriteCustomDirectory gdal_TIFFWriteCustomDirectory
#define TIFFWriteDirectory gdal_TIFFWriteDirectory
#define TIFFWriteDirectorySec gdal_TIFFWriteDirectorySec
#define TIFFWriteDirectoryTagAscii gdal_TIFFWriteDirectoryTagAscii
#define TIFFWriteDirectoryTagByteArray gdal_TIFFWriteDirectoryTagByteArray
#define TIFFWriteDirectoryTagCheckedAscii gdal_TIFFWriteDirectoryTagCheckedAscii
#define TIFFWriteDirectoryTagCheckedByteArray gdal_TIFFWriteDirectoryTagCheckedByteArray
#define TIFFWriteDirectoryTagCheckedDoubleArray gdal_TIFFWriteDirectoryTagCheckedDoubleArray
#define TIFFWriteDirectoryTagCheckedFloatArray gdal_TIFFWriteDirectoryTagCheckedFloatArray
#define TIFFWriteDirectoryTagCheckedIfd8Array gdal_TIFFWriteDirectoryTagCheckedIfd8Array
#define TIFFWriteDirectoryTagCheckedIfdArray gdal_TIFFWriteDirectoryTagCheckedIfdArray
#define TIFFWriteDirectoryTagCheckedLong gdal_TIFFWriteDirectoryTagCheckedLong
#define TIFFWriteDirectoryTagCheckedLong8Array gdal_TIFFWriteDirectoryTagCheckedLong8Array
#define TIFFWriteDirectoryTagCheckedLongArray gdal_TIFFWriteDirectoryTagCheckedLongArray
#define TIFFWriteDirectoryTagCheckedRational gdal_TIFFWriteDirectoryTagCheckedRational
#define TIFFWriteDirectoryTagCheckedRationalArray gdal_TIFFWriteDirectoryTagCheckedRationalArray
#define TIFFWriteDirectoryTagCheckedRationalDoubleArray gdal_TIFFWriteDirectoryTagCheckedRationalDoubleArray
#define TIFFWriteDirectoryTagCheckedSbyteArray gdal_TIFFWriteDirectoryTagCheckedSbyteArray
#define TIFFWriteDirectoryTagCheckedShort gdal_TIFFWriteDirectoryTagCheckedShort
#define TIFFWriteDirectoryTagCheckedShortArray gdal_TIFFWriteDirectoryTagCheckedShortArray
#define TIFFWriteDirectoryTagCheckedSlong8Array gdal_TIFFWriteDirectoryTagCheckedSlong8Array
#define TIFFWriteDirectoryTagCheckedSlongArray gdal_TIFFWriteDirectoryTagCheckedSlongArray
#define TIFFWriteDirectoryTagCheckedSrationalArray gdal_TIFFWriteDirectoryTagCheckedSrationalArray
#define TIFFWriteDirectoryTagCheckedSrationalDoubleArray gdal_TIFFWriteDirectoryTagCheckedSrationalDoubleArray
#define TIFFWriteDirectoryTagCheckedSshortArray gdal_TIFFWriteDirectoryTagCheckedSshortArray
#define TIFFWriteDirectoryTagCheckedUndefinedArray gdal_TIFFWriteDirectoryTagCheckedUndefinedArray
#define TIFFWriteDirectoryTagColormap gdal_TIFFWriteDirectoryTagColormap
#define TIFFWriteDirectoryTagData gdal_TIFFWriteDirectoryTagData
#define TIFFWriteDirectoryTagDoubleArray gdal_TIFFWriteDirectoryTagDoubleArray
#define TIFFWriteDirectoryTagFloatArray gdal_TIFFWriteDirectoryTagFloatArray
#define TIFFWriteDirectoryTagIfdArray gdal_TIFFWriteDirectoryTagIfdArray
#define TIFFWriteDirectoryTagIfdIfd8Array gdal_TIFFWriteDirectoryTagIfdIfd8Array
#define TIFFWriteDirectoryTagLong gdal_TIFFWriteDirectoryTagLong
#define TIFFWriteDirectoryTagLong8Array gdal_TIFFWriteDirectoryTagLong8Array
#define TIFFWriteDirectoryTagLongArray gdal_TIFFWriteDirectoryTagLongArray
#define TIFFWriteDirectoryTagLongLong8Array gdal_TIFFWriteDirectoryTagLongLong8Array
#define TIFFWriteDirectoryTagRational gdal_TIFFWriteDirectoryTagRational
#define TIFFWriteDirectoryTagRationalArray gdal_TIFFWriteDirectoryTagRationalArray
#define TIFFWriteDirectoryTagRationalDoubleArray gdal_TIFFWriteDirectoryTagRationalDoubleArray
#define TIFFWriteDirectoryTagSampleformatArray gdal_TIFFWriteDirectoryTagSampleformatArray
#define TIFFWriteDirectoryTagSbyteArray gdal_TIFFWriteDirectoryTagSbyteArray
#define TIFFWriteDirectoryTagShort gdal_TIFFWriteDirectoryTagShort
#define TIFFWriteDirectoryTagShortArray gdal_TIFFWriteDirectoryTagShortArray
#define TIFFWriteDirectoryTagShortLong gdal_TIFFWriteDirectoryTagShortLong
#define TIFFWriteDirectoryTagShortPerSample gdal_TIFFWriteDirectoryTagShortPerSample
#define TIFFWriteDirectoryTagSlong8Array gdal_TIFFWriteDirectoryTagSlong8Array
#define TIFFWriteDirectoryTagSlongArray gdal_TIFFWriteDirectoryTagSlongArray
#define TIFFWriteDirectoryTagSrationalArray gdal_TIFFWriteDirectoryTagSrationalArray
#define TIFFWriteDirectoryTagSrationalDoubleArray gdal_TIFFWriteDirectoryTagSrationalDoubleArray
#define TIFFWriteDirectoryTagSshortArray gdal_TIFFWriteDirectoryTagSshortArray
#define TIFFWriteDirectoryTagSubifd gdal_TIFFWriteDirectoryTagSubifd
#define TIFFWriteDirectoryTagTransferfunction gdal_TIFFWriteDirectoryTagTransferfunction
#define TIFFWriteDirectoryTagUndefinedArray gdal_TIFFWriteDirectoryTagUndefinedArray
#define TIFFWriteEncodedStrip gdal_TIFFWriteEncodedStrip
#define TIFFWriteEncodedTile gdal_TIFFWriteEncodedTile
#define _tiffWriteProc gdal__tiffWriteProc
#define TIFFWriteRawStrip gdal_TIFFWriteRawStrip
#define TIFFWriteRawTile gdal_TIFFWriteRawTile
#define TIFFWriteScanline gdal_TIFFWriteScanline
#define TIFFWriteTile gdal_TIFFWriteTile
#define TIFFXYZToRGB gdal_TIFFXYZToRGB
#define TIFFYCbCrtoRGB gdal_TIFFYCbCrtoRGB
#define TIFFYCbCrToRGBInit gdal_TIFFYCbCrToRGBInit
#define ToRationalEuclideanGCD gdal_ToRationalEuclideanGCD
#define TryChopUpUncompressedBigTiff gdal_TryChopUpUncompressedBigTiff
#define TWebPCleanup gdal_TWebPCleanup
#define TWebPDatasetWriter gdal_TWebPDatasetWriter
#define TWebPDecode gdal_TWebPDecode
#define TWebPEncode gdal_TWebPEncode
#define TWebPFixupTags gdal_TWebPFixupTags
#define TWebPPostEncode gdal_TWebPPostEncode
#define TWebPPreDecode gdal_TWebPPreDecode
#define TWebPPreEncode gdal_TWebPPreEncode
#define TWebPSetupDecode gdal_TWebPSetupDecode
#define TWebPSetupEncode gdal_TWebPSetupEncode
#define TWebPVGetField gdal_TWebPVGetField
#define TWebPVSetField gdal_TWebPVSetField
#define unixErrorHandler gdal_unixErrorHandler
#define unixWarningHandler gdal_unixWarningHandler
#define uv_decode gdal_uv_decode
#define uv_encode gdal_uv_encode
#define WriteAsLong4 gdal_WriteAsLong4
#define WriteAsLong8 gdal_WriteAsLong8
#define _WriteAsType gdal__WriteAsType
#define XYZtoRGB24 gdal_XYZtoRGB24
#define ZIPCleanup gdal_ZIPCleanup
#define ZIPDecode gdal_ZIPDecode
#define ZIPEncode gdal_ZIPEncode
#define ZIPFixupTags gdal_ZIPFixupTags
#define ZIPPostEncode gdal_ZIPPostEncode
#define ZIPPreDecode gdal_ZIPPreDecode
#define ZIPPreEncode gdal_ZIPPreEncode
#define ZIPSetupDecode gdal_ZIPSetupDecode
#define ZIPSetupEncode gdal_ZIPSetupEncode
#define ZIPVGetField gdal_ZIPVGetField
#define ZIPVSetField gdal_ZIPVSetField
#define ZSTDCleanup gdal_ZSTDCleanup
#define ZSTDDecode gdal_ZSTDDecode
#define ZSTDEncode gdal_ZSTDEncode
#define ZSTDFixupTags gdal_ZSTDFixupTags
#define ZSTDPostEncode gdal_ZSTDPostEncode
#define ZSTDPreDecode gdal_ZSTDPreDecode
#define ZSTDPreEncode gdal_ZSTDPreEncode
#define ZSTDSetupDecode gdal_ZSTDSetupDecode
#define ZSTDSetupEncode gdal_ZSTDSetupEncode
#define ZSTDVGetField gdal_ZSTDVGetField
#define ZSTDVSetField gdal_ZSTDVSetField
#define _msbmask gdal__msbmask
#define zeroruns gdal_zeroruns
#define oneruns gdal_oneruns
#define horizcode gdal_horizcode
#define passcode gdal_passcode
#define vcodes gdal_vcodes
#define photoTag gdal_photoTag
#define display_sRGB gdal_display_sRGB
#define anPrimes gdal_anPrimes
#define uv_row gdal_uv_row
#define TIFFBitRevTable gdal_TIFFBitRevTable
#define TIFFNoBitRevTable gdal_TIFFNoBitRevTable
#define twobitdeltas gdal_twobitdeltas
#define threebitdeltas gdal_threebitdeltas
#define TIFFVersion gdal_TIFFVersion
#define TIFFFaxBlackCodes gdal_TIFFFaxBlackCodes
#define TIFFFaxMainTable gdal_TIFFFaxMainTable
#define TIFFFaxWhiteTable gdal_TIFFFaxWhiteTable
#define TIFFFaxBlackTable gdal_TIFFFaxBlackTable
#define TIFFFaxWhiteCodes gdal_TIFFFaxWhiteCodes
#define tiffFieldArray gdal_tiffFieldArray
#define tiffFields gdal_tiffFields
#define exifFieldArray gdal_exifFieldArray
#define exifFields gdal_exifFields
#define gpsFieldArray gdal_gpsFieldArray
#define gpsFields gdal_gpsFields
#define faxFields gdal_faxFields
#define fax3Fields gdal_fax3Fields
#define fax4Fields gdal_fax4Fields
#define LERCFields gdal_LERCFields
#define LogLuvFields gdal_LogLuvFields
#define lzmaFields gdal_lzmaFields
#define ojpegFields gdal_ojpegFields
#define pixarlogFields gdal_pixarlogFields
#define predictFields gdal_predictFields
#define photoNames gdal_photoNames
#define orientNames gdal_orientNames
#define tagnames gdal_tagnames
#define TWebPFields gdal_TWebPFields
#define zipFields gdal_zipFields
#define ZSTDFields gdal_ZSTDFields
#define _TIFFBuiltinCODECS gdal__TIFFBuiltinCODECS
#define _TIFFerrorHandler gdal__TIFFerrorHandler
#define _TIFFwarningHandler gdal__TIFFwarningHandler
#define registeredCODECS gdal_registeredCODECS
#define _TIFFextender gdal__TIFFextender
#define Fltsize gdal_Fltsize
#define LogK1 gdal_LogK1
#define LogK2 gdal_LogK2
#define _TIFFwarningHandlerExt gdal__TIFFwarningHandlerExt
#define _TIFFerrorHandlerExt gdal__TIFFerrorHandlerExt

#define DISABLE_CHECK_TIFFSWABMACROS

#define TIFFInitDumpMode gdal_TIFFInitDumpMode
#ifdef LZW_SUPPORT
#define TIFFInitLZW gdal_TIFFInitLZW
#endif
#ifdef PACKBITS_SUPPORT
#define TIFFInitPackBits gdal_TIFFInitPackBits
#endif
#ifdef THUNDER_SUPPORT
#define TIFFInitThunderScan gdal_TIFFInitThunderScan
#endif
#ifdef NEXT_SUPPORT
#define TIFFInitNeXT gdal_TIFFInitNeXT
#endif
#ifdef JPEG_SUPPORT
#define TIFFInitJPEG gdal_TIFFInitJPEG
#define TIFFJPEGIsFullStripRequired gdal_TIFFJPEGIsFullStripRequired
#define TIFFReInitJPEG_12 gdal_TIFFReInitJPEG_12
#define TIFFJPEGIsFullStripRequired_12 gdal_TIFFJPEGIsFullStripRequired_12
#define TIFFInitJPEG_12 gdal_TIFFInitJPEG_12
#endif
#ifdef OJPEG_SUPPORT
#define TIFFInitOJPEG gdal_TIFFInitOJPEG
#endif
#ifdef CCITT_SUPPORT
#define TIFFInitCCITTRLE gdal_TIFFInitCCITTRLE
#define TIFFInitCCITTRLEW gdal_TIFFInitCCITTRLEW
#define TIFFInitCCITTFax3 gdal_TIFFInitCCITTFax3
#define TIFFInitCCITTFax4 gdal_TIFFInitCCITTFax4
#endif
#ifdef JBIG_SUPPORT
#define TIFFInitJBIG gdal_TIFFInitJBIG
#endif
#ifdef ZIP_SUPPORT
#define TIFFInitZIP gdal_TIFFInitZIP
#endif
#ifdef PIXARLOG_SUPPORT
#define TIFFInitPixarLog gdal_TIFFInitPixarLog
#endif
#ifdef LOGLUV_SUPPORT
#define TIFFInitSGILog gdal_TIFFInitSGILog
#endif
#ifdef LZMA_SUPPORT
#define TIFFInitLZMA gdal_TIFFInitLZMA
#endif
#ifdef WEBP_SUPPORT
#define TIFFInitWebP gdal_TIFFInitWebP
#endif
#ifdef LERC_SUPPORT
#define TIFFInitLERC gdal_TIFFInitLERC
#endif
#endif /* GDAL_LIBTIFF_SYMBOL_RENAME_H_INCLUDED */
