// SPDX-License-Identifier: MIT
// Copyright 2024, Even Rouault <even.rouault at spatialys.com>

#include "cpl_port.h"

CPL_C_START

const char *GMLASConfXSDGetFileContent(void);
const char *GMLASConfXMLGetFileContent(void);

CPL_C_END
