// File generated by generate_embedded_resources.py. DO NOT EDIT
{"gml_registry.xml", gml_registry_xml},
    {"inspire_cp_BasicPropertyUnit.gfs", inspire_cp_BasicPropertyUnit_gfs},
    {"inspire_cp_CadastralBoundary.gfs", inspire_cp_CadastralBoundary_gfs},
    {"inspire_cp_CadastralParcel.gfs", inspire_cp_CadastralParcel_gfs},
    {"inspire_cp_CadastralZoning.gfs", inspire_cp_CadastralZoning_gfs},
    {"jpfgdgml_AdmArea.gfs", jpfgdgml_AdmArea_gfs},
    {"jpfgdgml_AdmBdry.gfs", jpfgdgml_AdmBdry_gfs},
    {"jpfgdgml_AdmPt.gfs", jpfgdgml_AdmPt_gfs},
    {"jpfgdgml_BldA.gfs", jpfgdgml_BldA_gfs},
    {"jpfgdgml_BldL.gfs", jpfgdgml_BldL_gfs},
    {"jpfgdgml_Cntr.gfs", jpfgdgml_Cntr_gfs},
    {"jpfgdgml_CommBdry.gfs", jpfgdgml_CommBdry_gfs},
    {"jpfgdgml_CommPt.gfs", jpfgdgml_CommPt_gfs},
    {"jpfgdgml_Cstline.gfs", jpfgdgml_Cstline_gfs},
    {"jpfgdgml_ElevPt.gfs", jpfgdgml_ElevPt_gfs},
    {"jpfgdgml_GCP.gfs", jpfgdgml_GCP_gfs},
    {"jpfgdgml_LeveeEdge.gfs", jpfgdgml_LeveeEdge_gfs},
    {"jpfgdgml_RailCL.gfs", jpfgdgml_RailCL_gfs},
    {"jpfgdgml_RdASL.gfs", jpfgdgml_RdASL_gfs},
    {"jpfgdgml_RdArea.gfs", jpfgdgml_RdArea_gfs},
    {"jpfgdgml_RdCompt.gfs", jpfgdgml_RdCompt_gfs},
    {"jpfgdgml_RdEdg.gfs", jpfgdgml_RdEdg_gfs},
    {"jpfgdgml_RdMgtBdry.gfs", jpfgdgml_RdMgtBdry_gfs},
    {"jpfgdgml_RdSgmtA.gfs", jpfgdgml_RdSgmtA_gfs},
    {"jpfgdgml_RvrMgtBdry.gfs", jpfgdgml_RvrMgtBdry_gfs},
    {"jpfgdgml_SBAPt.gfs", jpfgdgml_SBAPt_gfs},
    {"jpfgdgml_SBArea.gfs", jpfgdgml_SBArea_gfs},
    {"jpfgdgml_SBBdry.gfs", jpfgdgml_SBBdry_gfs},
    {"jpfgdgml_WA.gfs", jpfgdgml_WA_gfs}, {"jpfgdgml_WL.gfs", jpfgdgml_WL_gfs},
    {"jpfgdgml_WStrA.gfs", jpfgdgml_WStrA_gfs},
    {"jpfgdgml_WStrL.gfs", jpfgdgml_WStrL_gfs},
    {"ruian_vf_ob_v1.gfs", ruian_vf_ob_v1_gfs},
    {"ruian_vf_st_uvoh_v1.gfs", ruian_vf_st_uvoh_v1_gfs},
    {"ruian_vf_st_v1.gfs", ruian_vf_st_v1_gfs},
    {"ruian_vf_v1.gfs", ruian_vf_v1_gfs},
