// File generated by generate_embedded_resources.py. DO NOT EDIT

static const char grib2_center_csv [] = {
    #embed "../../data/grib2_center.csv"
            , 0
};

static const char grib2_process_csv [] = {
    #embed "../../data/grib2_process.csv"
            , 0
};

static const char grib2_subcenter_csv [] = {
    #embed "../../data/grib2_subcenter.csv"
            , 0
};

static const char grib2_table_4_2_0_0_csv [] = {
    #embed "../../data/grib2_table_4_2_0_0.csv"
            , 0
};

static const char grib2_table_4_2_0_1_csv [] = {
    #embed "../../data/grib2_table_4_2_0_1.csv"
            , 0
};

static const char grib2_table_4_2_0_13_csv [] = {
    #embed "../../data/grib2_table_4_2_0_13.csv"
            , 0
};

static const char grib2_table_4_2_0_14_csv [] = {
    #embed "../../data/grib2_table_4_2_0_14.csv"
            , 0
};

static const char grib2_table_4_2_0_15_csv [] = {
    #embed "../../data/grib2_table_4_2_0_15.csv"
            , 0
};

static const char grib2_table_4_2_0_16_csv [] = {
    #embed "../../data/grib2_table_4_2_0_16.csv"
            , 0
};

static const char grib2_table_4_2_0_17_csv [] = {
    #embed "../../data/grib2_table_4_2_0_17.csv"
            , 0
};

static const char grib2_table_4_2_0_18_csv [] = {
    #embed "../../data/grib2_table_4_2_0_18.csv"
            , 0
};

static const char grib2_table_4_2_0_19_csv [] = {
    #embed "../../data/grib2_table_4_2_0_19.csv"
            , 0
};

static const char grib2_table_4_2_0_190_csv [] = {
    #embed "../../data/grib2_table_4_2_0_190.csv"
            , 0
};

static const char grib2_table_4_2_0_191_csv [] = {
    #embed "../../data/grib2_table_4_2_0_191.csv"
            , 0
};

static const char grib2_table_4_2_0_2_csv [] = {
    #embed "../../data/grib2_table_4_2_0_2.csv"
            , 0
};

static const char grib2_table_4_2_0_20_csv [] = {
    #embed "../../data/grib2_table_4_2_0_20.csv"
            , 0
};

static const char grib2_table_4_2_0_21_csv [] = {
    #embed "../../data/grib2_table_4_2_0_21.csv"
            , 0
};

static const char grib2_table_4_2_0_22_csv [] = {
    #embed "../../data/grib2_table_4_2_0_22.csv"
            , 0
};

static const char grib2_table_4_2_0_3_csv [] = {
    #embed "../../data/grib2_table_4_2_0_3.csv"
            , 0
};

static const char grib2_table_4_2_0_4_csv [] = {
    #embed "../../data/grib2_table_4_2_0_4.csv"
            , 0
};

static const char grib2_table_4_2_0_5_csv [] = {
    #embed "../../data/grib2_table_4_2_0_5.csv"
            , 0
};

static const char grib2_table_4_2_0_6_csv [] = {
    #embed "../../data/grib2_table_4_2_0_6.csv"
            , 0
};

static const char grib2_table_4_2_0_7_csv [] = {
    #embed "../../data/grib2_table_4_2_0_7.csv"
            , 0
};

static const char grib2_table_4_2_10_0_csv [] = {
    #embed "../../data/grib2_table_4_2_10_0.csv"
            , 0
};

static const char grib2_table_4_2_10_1_csv [] = {
    #embed "../../data/grib2_table_4_2_10_1.csv"
            , 0
};

static const char grib2_table_4_2_10_191_csv [] = {
    #embed "../../data/grib2_table_4_2_10_191.csv"
            , 0
};

static const char grib2_table_4_2_10_2_csv [] = {
    #embed "../../data/grib2_table_4_2_10_2.csv"
            , 0
};

static const char grib2_table_4_2_10_3_csv [] = {
    #embed "../../data/grib2_table_4_2_10_3.csv"
            , 0
};

static const char grib2_table_4_2_10_4_csv [] = {
    #embed "../../data/grib2_table_4_2_10_4.csv"
            , 0
};

static const char grib2_table_4_2_1_0_csv [] = {
    #embed "../../data/grib2_table_4_2_1_0.csv"
            , 0
};

static const char grib2_table_4_2_1_1_csv [] = {
    #embed "../../data/grib2_table_4_2_1_1.csv"
            , 0
};

static const char grib2_table_4_2_1_2_csv [] = {
    #embed "../../data/grib2_table_4_2_1_2.csv"
            , 0
};

static const char grib2_table_4_2_20_0_csv [] = {
    #embed "../../data/grib2_table_4_2_20_0.csv"
            , 0
};

static const char grib2_table_4_2_20_1_csv [] = {
    #embed "../../data/grib2_table_4_2_20_1.csv"
            , 0
};

static const char grib2_table_4_2_20_2_csv [] = {
    #embed "../../data/grib2_table_4_2_20_2.csv"
            , 0
};

static const char grib2_table_4_2_20_3_csv [] = {
    #embed "../../data/grib2_table_4_2_20_3.csv"
            , 0
};

static const char grib2_table_4_2_2_0_csv [] = {
    #embed "../../data/grib2_table_4_2_2_0.csv"
            , 0
};

static const char grib2_table_4_2_2_3_csv [] = {
    #embed "../../data/grib2_table_4_2_2_3.csv"
            , 0
};

static const char grib2_table_4_2_2_4_csv [] = {
    #embed "../../data/grib2_table_4_2_2_4.csv"
            , 0
};

static const char grib2_table_4_2_2_5_csv [] = {
    #embed "../../data/grib2_table_4_2_2_5.csv"
            , 0
};

static const char grib2_table_4_2_2_6_csv [] = {
    #embed "../../data/grib2_table_4_2_2_6.csv"
            , 0
};

static const char grib2_table_4_2_2_7_csv [] = {
    #embed "../../data/grib2_table_4_2_2_7.csv"
            , 0
};

static const char grib2_table_4_2_3_0_csv [] = {
    #embed "../../data/grib2_table_4_2_3_0.csv"
            , 0
};

static const char grib2_table_4_2_3_1_csv [] = {
    #embed "../../data/grib2_table_4_2_3_1.csv"
            , 0
};

static const char grib2_table_4_2_3_2_csv [] = {
    #embed "../../data/grib2_table_4_2_3_2.csv"
            , 0
};

static const char grib2_table_4_2_3_3_csv [] = {
    #embed "../../data/grib2_table_4_2_3_3.csv"
            , 0
};

static const char grib2_table_4_2_3_4_csv [] = {
    #embed "../../data/grib2_table_4_2_3_4.csv"
            , 0
};

static const char grib2_table_4_2_3_5_csv [] = {
    #embed "../../data/grib2_table_4_2_3_5.csv"
            , 0
};

static const char grib2_table_4_2_3_6_csv [] = {
    #embed "../../data/grib2_table_4_2_3_6.csv"
            , 0
};

static const char grib2_table_4_2_4_0_csv [] = {
    #embed "../../data/grib2_table_4_2_4_0.csv"
            , 0
};

static const char grib2_table_4_2_4_1_csv [] = {
    #embed "../../data/grib2_table_4_2_4_1.csv"
            , 0
};

static const char grib2_table_4_2_4_10_csv [] = {
    #embed "../../data/grib2_table_4_2_4_10.csv"
            , 0
};

static const char grib2_table_4_2_4_2_csv [] = {
    #embed "../../data/grib2_table_4_2_4_2.csv"
            , 0
};

static const char grib2_table_4_2_4_3_csv [] = {
    #embed "../../data/grib2_table_4_2_4_3.csv"
            , 0
};

static const char grib2_table_4_2_4_4_csv [] = {
    #embed "../../data/grib2_table_4_2_4_4.csv"
            , 0
};

static const char grib2_table_4_2_4_5_csv [] = {
    #embed "../../data/grib2_table_4_2_4_5.csv"
            , 0
};

static const char grib2_table_4_2_4_6_csv [] = {
    #embed "../../data/grib2_table_4_2_4_6.csv"
            , 0
};

static const char grib2_table_4_2_4_7_csv [] = {
    #embed "../../data/grib2_table_4_2_4_7.csv"
            , 0
};

static const char grib2_table_4_2_4_8_csv [] = {
    #embed "../../data/grib2_table_4_2_4_8.csv"
            , 0
};

static const char grib2_table_4_2_4_9_csv [] = {
    #embed "../../data/grib2_table_4_2_4_9.csv"
            , 0
};

static const char grib2_table_4_2_local_Canada_csv [] = {
    #embed "../../data/grib2_table_4_2_local_Canada.csv"
            , 0
};

static const char grib2_table_4_2_local_HPC_csv [] = {
    #embed "../../data/grib2_table_4_2_local_HPC.csv"
            , 0
};

static const char grib2_table_4_2_local_MRMS_csv [] = {
    #embed "../../data/grib2_table_4_2_local_MRMS.csv"
            , 0
};

static const char grib2_table_4_2_local_NCEP_csv [] = {
    #embed "../../data/grib2_table_4_2_local_NCEP.csv"
            , 0
};

static const char grib2_table_4_2_local_NDFD_csv [] = {
    #embed "../../data/grib2_table_4_2_local_NDFD.csv"
            , 0
};

static const char grib2_table_4_2_local_index_csv [] = {
    #embed "../../data/grib2_table_4_2_local_index.csv"
            , 0
};

static const char grib2_table_4_5_csv [] = {
    #embed "../../data/grib2_table_4_5.csv"
            , 0
};

static const char grib2_table_versions_csv [] = {
    #embed "../../data/grib2_table_versions.csv"
            , 0
};
