/******************************************************************************
 *
 * Project:  XLS Translator
 * Purpose:  Include freexl.h
 * Author:   Even Rouault, even dot rouault at spatialys.com
 *
 ******************************************************************************
 * Copyright (c) 2011-2012, Even Rouault <even dot rouault at spatialys.com>
 *
 * SPDX-License-Identifier: MIT
 ****************************************************************************/

#ifndef INCLUDE_FREEXL_H_INCLUDED
#define INCLUDE_FREEXL_H_INCLUDED

#ifdef __clang__
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdocumentation"
#endif

#include <freexl.h>

#ifdef __clang
#pragma clang diagnostic pop
#endif

#endif  // INCLUDE_FREEXL_H_INCLUDED
