// File generated by generate_embedded_resources.py. DO NOT EDIT

static const char gml_registry_xml[] = {
#embed "data/gml_registry.xml"
    , 0};

static const char inspire_cp_BasicPropertyUnit_gfs[] = {
#embed "data/inspire_cp_BasicPropertyUnit.gfs"
    , 0};

static const char inspire_cp_CadastralBoundary_gfs[] = {
#embed "data/inspire_cp_CadastralBoundary.gfs"
    , 0};

static const char inspire_cp_CadastralParcel_gfs[] = {
#embed "data/inspire_cp_CadastralParcel.gfs"
    , 0};

static const char inspire_cp_CadastralZoning_gfs[] = {
#embed "data/inspire_cp_CadastralZoning.gfs"
    , 0};

static const char jpfgdgml_AdmArea_gfs[] = {
#embed "data/jpfgdgml_AdmArea.gfs"
    , 0};

static const char jpfgdgml_AdmBdry_gfs[] = {
#embed "data/jpfgdgml_AdmBdry.gfs"
    , 0};

static const char jpfgdgml_AdmPt_gfs[] = {
#embed "data/jpfgdgml_AdmPt.gfs"
    , 0};

static const char jpfgdgml_BldA_gfs[] = {
#embed "data/jpfgdgml_BldA.gfs"
    , 0};

static const char jpfgdgml_BldL_gfs[] = {
#embed "data/jpfgdgml_BldL.gfs"
    , 0};

static const char jpfgdgml_Cntr_gfs[] = {
#embed "data/jpfgdgml_Cntr.gfs"
    , 0};

static const char jpfgdgml_CommBdry_gfs[] = {
#embed "data/jpfgdgml_CommBdry.gfs"
    , 0};

static const char jpfgdgml_CommPt_gfs[] = {
#embed "data/jpfgdgml_CommPt.gfs"
    , 0};

static const char jpfgdgml_Cstline_gfs[] = {
#embed "data/jpfgdgml_Cstline.gfs"
    , 0};

static const char jpfgdgml_ElevPt_gfs[] = {
#embed "data/jpfgdgml_ElevPt.gfs"
    , 0};

static const char jpfgdgml_GCP_gfs[] = {
#embed "data/jpfgdgml_GCP.gfs"
    , 0};

static const char jpfgdgml_LeveeEdge_gfs[] = {
#embed "data/jpfgdgml_LeveeEdge.gfs"
    , 0};

static const char jpfgdgml_RailCL_gfs[] = {
#embed "data/jpfgdgml_RailCL.gfs"
    , 0};

static const char jpfgdgml_RdASL_gfs[] = {
#embed "data/jpfgdgml_RdASL.gfs"
    , 0};

static const char jpfgdgml_RdArea_gfs[] = {
#embed "data/jpfgdgml_RdArea.gfs"
    , 0};

static const char jpfgdgml_RdCompt_gfs[] = {
#embed "data/jpfgdgml_RdCompt.gfs"
    , 0};

static const char jpfgdgml_RdEdg_gfs[] = {
#embed "data/jpfgdgml_RdEdg.gfs"
    , 0};

static const char jpfgdgml_RdMgtBdry_gfs[] = {
#embed "data/jpfgdgml_RdMgtBdry.gfs"
    , 0};

static const char jpfgdgml_RdSgmtA_gfs[] = {
#embed "data/jpfgdgml_RdSgmtA.gfs"
    , 0};

static const char jpfgdgml_RvrMgtBdry_gfs[] = {
#embed "data/jpfgdgml_RvrMgtBdry.gfs"
    , 0};

static const char jpfgdgml_SBAPt_gfs[] = {
#embed "data/jpfgdgml_SBAPt.gfs"
    , 0};

static const char jpfgdgml_SBArea_gfs[] = {
#embed "data/jpfgdgml_SBArea.gfs"
    , 0};

static const char jpfgdgml_SBBdry_gfs[] = {
#embed "data/jpfgdgml_SBBdry.gfs"
    , 0};

static const char jpfgdgml_WA_gfs[] = {
#embed "data/jpfgdgml_WA.gfs"
    , 0};

static const char jpfgdgml_WL_gfs[] = {
#embed "data/jpfgdgml_WL.gfs"
    , 0};

static const char jpfgdgml_WStrA_gfs[] = {
#embed "data/jpfgdgml_WStrA.gfs"
    , 0};

static const char jpfgdgml_WStrL_gfs[] = {
#embed "data/jpfgdgml_WStrL.gfs"
    , 0};

static const char ruian_vf_ob_v1_gfs[] = {
#embed "data/ruian_vf_ob_v1.gfs"
    , 0};

static const char ruian_vf_st_uvoh_v1_gfs[] = {
#embed "data/ruian_vf_st_uvoh_v1.gfs"
    , 0};

static const char ruian_vf_st_v1_gfs[] = {
#embed "data/ruian_vf_st_v1.gfs"
    , 0};

static const char ruian_vf_v1_gfs[] = {
#embed "data/ruian_vf_v1.gfs"
    , 0};
