/******************************************************************************
 *
 * Project:  GDAL
 * Purpose:  Includes libwebp headers
 * Author:   Even Rouault <even dot rouault at spatialys dot com>
 *
 ******************************************************************************
 * Copyright (c) 2016, Even Rouault <even dot rouault at spatialys dot com>
 *
 * SPDX-License-Identifier: MIT
 *****************************************************************************/

#ifndef GDAL_LIBWEBP_HEADERS_H
#define GDAL_LIBWEBP_HEADERS_H

#include "cpl_port.h"

#ifdef HAVE_GCC_SYSTEM_HEADER
#pragma GCC system_header
#endif

#include "webp/decode.h"
#include "webp/encode.h"

#endif /* GDAL_LIBWEBP_HEADERS_H */
